# ------------------------------------------------------------------------------
# Test Calendar Page
# ------------------------------------------------------------------------------
import sys
import datetime as dt
from django_bs_test import TestCase
from django.contrib.auth.models import User
from django.utils import translation
from wagtail.core.models import Page
from ls.joyous.models.calendar import CalendarPage
from ls.joyous.models.events import SimpleEventPage

# ------------------------------------------------------------------------------
class Test(TestCase):
    def setUp(self):
        self.user = User.objects.create_user('i', 'i@j.test', 's3(r3t')
        calendar = CalendarPage(owner  = self.user,
                                slug  = "events",
                                title = "Events")
        Page.objects.get(slug='home').add_child(instance=calendar)
        calendar.save_revision().publish()
        event = SimpleEventPage(owner = self.user,
                                slug  = "tree-planting",
                                title = "Tree Planting",
                                date      = dt.date(2011,6,5),
                                time_from = dt.time(9,30),
                                time_to   = dt.time(11,0))
        calendar.add_child(instance=event)
        event.save_revision().publish()

    def testMonthView(self):
        response = self.client.get("/events/2012/03/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(select('thead a')), 4)
        self.assertEqual(len(select("table.calendar thead tr th.sun")), 1)
        month = select("tr.heading th.month .month-name")[0]
        self.assertEqual(month.string.strip(), "March")
        self.assertEqual(len(select("tbody tr")), 5)
        self.assertEqual(len(select("tbody td")), 35)
        self.assertEqual(len(select("tbody td.day")), 31)
        self.assertEqual(len(select("tbody td.noday")), 4)
        holidayNames = [holiday.string.strip() for holiday in
                        select("tbody td.day .holiday .holiday-name")]
        self.assertEqual(holidayNames,
                         ["Taranaki Anniversary Day", "Otago Anniversary Day"])

    def testMonthAbbr(self):
        response = self.client.get("/events/2012/Apr/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(select("table.calendar thead tr th.sun")), 1)
        month = select("tr.heading th.month .month-name")[0]
        self.assertEqual(month.string.strip(), "April")

    def testWeekView(self):
        response = self.client.get("/events/2012/W11/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(select('thead a')), 4)
        self.assertEqual(len(select("table.calendar thead tr th.sun")), 1)
        self.assertEqual(len(select("tbody tr")), 1)
        self.assertEqual(len(select("tbody td.day")), 7)
        holidays = select("tbody td.day .holiday")
        self.assertEqual(len(holidays), 1)
        self.assertEqual(holidays[0].h4.string.strip(), "12 Mar")
        self.assertEqual(holidays[0].div.string.strip(),
                         "Taranaki Anniversary Day")

    def testDayWithOneEvent(self):
        response = self.client.get("/events/2011/6/5/")
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/events/tree-planting/")

    def testUpcomingEvents(self):
        response = self.client.get("/events/upcoming/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(select(".upcoming-events")), 1)
        self.assertEqual(len(select(".upcoming-events .event-item")), 0)

    def testPastEvents(self):
        response = self.client.get("/events/past/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(select(".past-events")), 1)
        events = select(".past-events .event-item")
        self.assertEqual(len(events), 1)
        title = events[0].select("a.event-title")[0]
        self.assertEqual(title.string.strip(), "Tree Planting")
        self.assertEqual(title['href'], "/events/tree-planting/")
        when = events[0].select(".event-when")[0]
        self.assertEqual(when.string.strip(),
                         "Sunday 5th of June 2011 at 9:30am to 11am")

    def testMiniMonthView(self):
        response = self.client.get("/events/mini/2011/06/")
        self.assertEqual(response.status_code, 404)

    def testMiniMonthAjaxView(self):
        response = self.client.get("/events/mini/2011/06/",
                                   HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(select("table.minicalendar thead tr th.sun")), 1)
        month = select("tr.heading th.month .month-name")[0]
        self.assertEqual(month.string.strip(), "June")
        self.assertEqual(len(select("tbody tr")), 5)
        self.assertEqual(len(select("tbody td")), 35)
        self.assertEqual(len(select("tbody td.day")), 30)
        self.assertEqual(len(select("tbody td.noday")), 5)
        event = select("tbody td.day a.event")[0]
        self.assertEqual(event.string.strip(), "5")
        self.assertEqual(event['href'], "/events/2011/06/05/")
        self.assertEqual(event['title'], "Tree Planting")

    def testInvalidDates(self):
        invalidDates = ["2012/13", "2008/W54", "2099/W53"]
        for date in invalidDates:
            with self.subTest(date=date):
                response = self.client.get("/events/{}/".format(date))
                self.assertEqual(response.status_code, 404)

    def testCalendarStartMonth(self):
        response = self.client.get("/events/1900/1/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        links = select('thead a')
        self.assertEqual(len(links), 2)
        self.assertEqual(links[0].get('title'), "Next month")
        self.assertEqual(links[0].get('href'), "/events/1900/2/")
        self.assertEqual(links[1].get('title'), "Next year")
        self.assertEqual(links[1].get('href'), "/events/1901/1/")

    def testCalendarFinishMonth(self):
        response = self.client.get("/events/2099/12/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        links = select('thead a')
        self.assertEqual(len(links), 2)
        self.assertEqual(links[0].get('title'), "Previous month")
        self.assertEqual(links[0].get('href'), "/events/2099/11/")
        self.assertEqual(links[1].get('title'), "Previous year")
        self.assertEqual(links[1].get('href'), "/events/2098/12/")

    def testCalendarStartWeek(self):
        response = self.client.get("/events/1900/W1/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        links0 = select('.calendar-options .events-view a')
        self.assertEqual(len(links0), 3)
        self.assertEqual(links0[0].get_text(), "List View")
        self.assertEqual(links0[1].get_text(), "This Week")
        self.assertEqual(links0[2].get_text(), "Monthly View")
        self.assertEqual(links0[2].get('href'), "/events/1900/1/")
        links1 = select('thead a')
        self.assertEqual(len(links1), 2)
        self.assertEqual(links1[0].get('title'), "Next week")
        self.assertEqual(links1[0].get('href'), "/events/1900/W2/")
        self.assertEqual(links1[1].get('title'), "Next year")
        self.assertEqual(links1[1].get('href'), "/events/1901/W1/")

    def testCalendarFinishWeek(self):
        response = self.client.get("/events/2099/W52/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        links0 = select('.calendar-options .events-view a')
        self.assertEqual(len(links0), 3)
        self.assertEqual(links0[0].get_text(), "List View")
        self.assertEqual(links0[1].get_text(), "This Week")
        self.assertEqual(links0[2].get_text(), "Monthly View")
        self.assertEqual(links0[2].get('href'), "/events/2099/12/")
        links1 = select('thead a')
        self.assertEqual(len(links1), 2)
        self.assertEqual(links1[0].get('title'), "Previous week")
        self.assertEqual(links1[0].get('href'), "/events/2099/W51/")
        self.assertEqual(links1[1].get('title'), "Previous year")
        self.assertEqual(links1[1].get('href'), "/events/2098/W52/")

    def testNextYearW53(self):
        response = self.client.get("/events/2098/W53/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        links = select('thead a')
        self.assertEqual(len(links), 3)
        self.assertEqual(links[0].get('title'), "Previous week")
        self.assertEqual(links[0].get('href'), "/events/2098/W52/")
        self.assertEqual(links[1].get('title'), "Next week")
        self.assertEqual(links[1].get('href'), "/events/2099/W1/")
        self.assertEqual(links[2].get('title'), "Previous year")
        self.assertEqual(links[2].get('href'), "/events/2097/W52/")

    def testFutureDates(self):
        response = self.client.get("/events/2525/1/")
        self.assertEqual(response.status_code, 404)
        response = self.client.get("/events/2100/W1/")
        self.assertEqual(response.status_code, 404)

# ------------------------------------------------------------------------------
class TestFrançais(TestCase):
    def setUp(self):
        translation.activate('fr')
        self.user = User.objects.create_superuser('i', 'i@j.test', 's3(r3t')
        calendar = CalendarPage(owner  = self.user,
                                slug  = "calendrier",
                                title = "Calendrier")
        Page.objects.get(slug='home').add_child(instance=calendar)
        calendar.save_revision().publish()
        event = SimpleEventPage(owner = self.user,
                                slug  = "plantation-d-arbres",
                                title = "Plantation d'arbres",
                                date      = dt.date(2011,6,5),
                                time_from = dt.time(9,30),
                                time_to   = dt.time(11,0))
        calendar.add_child(instance=event)
        event.save_revision().publish()

    def tearDown(self):
        translation.deactivate()

    def testMonthView(self):
        response = self.client.get("/calendrier/2012/03/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        month = select("tr.heading th.month .month-name")[0]
        self.assertEqual(month.string.strip(), "Mars")

    def testWeekView(self):
        response = self.client.get("/calendrier/2012/W11/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        holidays = select("tbody td.day .holiday")
        self.assertEqual(holidays[0].h4.string.strip(), "12 Mar")
        self.assertEqual(holidays[0].div.string.strip(),
                         "Taranaki Anniversary Day")

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Date/time utilities
# ------------------------------------------------------------------------------
import datetime as dt
from functools import wraps
from inspect import signature
from django.conf import settings
from django.utils import dateformat
from django.utils import timezone
from django.utils.translation import gettext as _

# ------------------------------------------------------------------------------
def getLocalDate(*args, **kwargs):
    """
    Get the date in the local timezone from date and optionally time
    """
    return getLocalDateAndTime(*args, **kwargs)[0]

def getLocalTime(date, time, *args, **kwargs):
    """
    Get the time in the local timezone from date and time
    """
    if time is not None:
        return getLocalDateAndTime(date, time, *args, **kwargs)[1]

def getLocalDateAndTime(date, time, *args, **kwargs):
    """
    Get the date and time in the local timezone from date and optionally time
    """
    localDt = getLocalDatetime(date, time, *args, **kwargs)
    if time is not None:
        return (localDt.date(), localDt.timetz())
    else:
        return (localDt.date(), None)

def getLocalDatetime(date, time, tz=None, timeDefault=dt.time.max):
    """
    Get a datetime in the local timezone from date and optionally time
    """
    localTZ = timezone.get_current_timezone()
    if tz is None or tz == localTZ:
        localDt = getAwareDatetime(date, time, tz, timeDefault)
    else:
        # create in event's time zone
        eventDt = getAwareDatetime(date, time, tz, timeDefault)
        # convert to local time zone
        localDt = eventDt.astimezone(localTZ)
        if time is None:
            localDt = getAwareDatetime(localDt.date(), None, localTZ, timeDefault)
    return localDt

def getAwareDatetime(date, time, tz, timeDefault=dt.time.max):
    """
    Get a datetime in the given timezone from date and optionally time.
    If time is not given it will default to timeDefault if that is given
    or if not then to the end of the day.
    """
    if time is None:
        time = timeDefault
    datetime = dt.datetime.combine(date, time)
    # arbitary rule to handle DST transitions:
    # if daylight savings causes an error then use standard time
    datetime = timezone.make_aware(datetime, tz, is_dst=False)
    return datetime

def todayUtc():
    """
    The current date in the UTC timezone
    """
    return dt.datetime.utcnow().date()

# ------------------------------------------------------------------------------
def timeFrom(time_from):
    """
    Return time_from if it is set, otherwise return the start of the day
    """
    return time_from if time_from is not None else dt.time.min

def timeTo(time_to):
    """
    Return time_to if it is set, otherwise return the end of the day
    """
    return time_to if time_to is not None else dt.time.max

# ------------------------------------------------------------------------------
def timeFormat(time_from, time_to=None, prefix="", infix=None):
    """
    Format the times time_from and optionally time_to, e.g. 10am
    """
    retval = ""
    if time_from != "" and time_from is not None:
        retval += prefix
        retval += dateformat.time_format(time_from, "fA").lower()
    if time_to != "" and time_to is not None:
        to = format(dateformat.time_format(time_to, "fA").lower())
        if infix is not None:
            retval = "{} {} {}".format(retval, infix, to)
        else:
            retval = _("{fromTime} to {toTime}").format(fromTime=retval,
                                                        toTime=to)
    return retval.strip()

def dateFormat(when):
    """
    Format the date when, e.g. Friday 14th of April 2011
    """
    retval = ""
    if when is not None:
        dow = dateformat.format(when, "l")
        dom = dateformat.format(when, "jS")
        month = dateformat.format(when, "F")
        if when.year != dt.date.today().year:
            retval = _("{weekday} {day} of {month} {year}")  \
                    .format(weekday=dow, day=dom, month=month, year=when.year)
        else:
            retval = _("{weekday} {day} of {month}")  \
                    .format(weekday=dow, day=dom, month=month)
    return retval

def dateFormatDMY(when):
    """
    Format when as day month year, e.g. 14 April 2017
    """
    if when is not None:
        return dateformat.format(when, "j F Y")
    else:
        return ""

# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Test Many Things Utilities
# ------------------------------------------------------------------------------
import sys
import datetime as dt
import pytz
from django.test import TestCase
from django.utils import translation
from ls.joyous.utils.manythings import toOrdinal, toTheOrdinal, hrJoin

# ------------------------------------------------------------------------------
class Test(TestCase):
    def testToOrdinal(self):
        self.assertEqual(toOrdinal(-1), "last")
        self.assertEqual(toOrdinal(-2), "penultimate")
        self.assertEqual(toOrdinal(1), "first")
        self.assertEqual(toOrdinal(2), "second")
        self.assertEqual(toOrdinal(3), "third")
        self.assertEqual(toOrdinal(4), "fourth")
        self.assertEqual(toOrdinal(5), "fifth")

    def testToOrdinalNum(self):
        self.assertEqual(toOrdinal(6), "6th")
        self.assertEqual(toOrdinal(11), "11th")
        self.assertEqual(toOrdinal(12), "12th")
        self.assertEqual(toOrdinal(13), "13th")
        self.assertEqual(toOrdinal(21), "21st")
        self.assertEqual(toOrdinal(102), "102nd")
        self.assertEqual(toOrdinal(6543), "6543rd")

    def testToTheOrdinal(self):
        self.assertEqual(toTheOrdinal(-1), "The Last")
        self.assertEqual(toTheOrdinal(-2, False), "the penultimate")
        self.assertEqual(toTheOrdinal(1), "The First")
        self.assertEqual(toTheOrdinal(2), "The Second")
        self.assertEqual(toTheOrdinal(3), "The Third")
        self.assertEqual(toTheOrdinal(4), "The Fourth")
        self.assertEqual(toTheOrdinal(5), "The Fifth")

    def testToTheOrdinalNum(self):
        self.assertEqual(toTheOrdinal(6), "The 6th")
        self.assertEqual(toTheOrdinal(11), "The 11th")
        self.assertEqual(toTheOrdinal(12), "The 12th")
        self.assertEqual(toTheOrdinal(13), "The 13th")
        self.assertEqual(toTheOrdinal(21), "The 21st")
        self.assertEqual(toTheOrdinal(102), "The 102nd")
        self.assertEqual(toTheOrdinal(6543), "The 6543rd")

    def testHumanReadableJoin(self):
        self.assertEqual(hrJoin([""]), "")
        self.assertEqual(hrJoin(["ice"]), "ice")
        self.assertEqual(hrJoin(["ice", "fire"]), "ice and fire")
        self.assertEqual(hrJoin(["wind", "ice", "fire"]),
                         "wind, ice and fire")
        self.assertEqual(hrJoin(["dog", "cat", "hen", "yak", "ant"]),
                         "dog, cat, hen, yak and ant")

# ------------------------------------------------------------------------------
class  TestFrançais(TestCase):
    def setUp(self):
        translation.activate('fr')

    def tearDown(self):
        translation.deactivate()

    def testToOrdinal(self):
        self.assertEqual(toOrdinal(-1), "dernier")
        self.assertEqual(toOrdinal(-2), "avant-dernier")
        self.assertEqual(toOrdinal (1), "premier")
        self.assertEqual(toOrdinal (2), "deuxième")
        self.assertEqual(toOrdinal (3), "troisième")
        self.assertEqual(toOrdinal (4), "quatrième")
        self.assertEqual(toOrdinal (5), "cinquième")

    def testToOrdinalNum(self):
        self.assertEqual(toOrdinal(6), "6me")
        self.assertEqual(toOrdinal(11), "11er")
        self.assertEqual(toOrdinal(12), "12me")
        self.assertEqual(toOrdinal(13), "13me")
        self.assertEqual(toOrdinal(21), "21er")
        self.assertEqual(toOrdinal(102), "102me")
        self.assertEqual(toOrdinal(6543), "6543me")

    def testToTheOrdinal(self):
        self.assertEqual(toTheOrdinal(-1), "Le Dernier")
        self.assertEqual(toTheOrdinal(-2, True), "L'Avant-Dernier")
        self.assertEqual(toTheOrdinal(-2, False), "l'avant-dernier")
        self.assertEqual(toTheOrdinal(1), "La Premier")
        self.assertEqual(toTheOrdinal(2, False), "la deuxième")
        self.assertEqual(toTheOrdinal(3), "Le Troisième")
        self.assertEqual(toTheOrdinal(4), "Le Quatrième")
        self.assertEqual(toTheOrdinal(5), "Le Cinquième")

    def testToOrdinalNum(self):
        self.assertEqual(toTheOrdinal(6), "La 6me")
        self.assertEqual(toTheOrdinal(11), "La 11er")
        self.assertEqual(toTheOrdinal(12), "La 12me")
        self.assertEqual(toTheOrdinal(13), "La 13me")
        self.assertEqual(toTheOrdinal(21), "La 21er")
        self.assertEqual(toTheOrdinal(102), "La 102me")
        self.assertEqual(toTheOrdinal(6543), "La 6543me")

    def testHumanReadableJoin(self):
        self.assertEqual(hrJoin([""]), "")
        self.assertEqual(hrJoin (["glace"]), "glace")
        self.assertEqual(hrJoin (["glace", "feu"]), "glace et feu")
        self.assertEqual(hrJoin (["vent", "glace", "feu"]),
                         "vent, glace et feu")
        self.assertEqual(hrJoin (["chien", "chat", "poule", "yak", "fourmi"]),
                         "chien, chat, poule, yak et fourmi")

# ------------------------------------------------------------------------------
class  TestΕλληνικά(TestCase):
    def setUp(self):
        translation.activate('el')

    def tearDown(self):
        translation.deactivate()

    def testToOrdinal(self):
        self.assertEqual(toOrdinal(-1), "τελευταίος")
        self.assertEqual(toOrdinal(-2), "προτελευταία")
        self.assertEqual(toOrdinal (1), "τελευταίο")
        self.assertEqual(toOrdinal (2), "προτελευταία")
        self.assertEqual(toOrdinal (3), "πρώτη")
        self.assertEqual(toOrdinal (4), "δεύτερη")
        self.assertEqual(toOrdinal (5), "τρίτη")

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# ical import/export format
# ------------------------------------------------------------------------------
import datetime as dt
from collections import namedtuple
from contextlib import suppress
from itertools import chain
import pytz
import base64
import quopri
from icalendar import Calendar, Event
from icalendar import vDatetime, vRecur, vDDDTypes, vText
from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponse
from django.utils import timezone
from ls.joyous import __version__
from ..models import (SimpleEventPage, MultidayEventPage, RecurringEventPage,
        MultidayRecurringEventPage, EventExceptionBase, ExtraInfoPage,
        CancellationPage, PostponementPage, RescheduleMultidayEventPage,
        EventBase, CalendarPage)
from ..utils.recurrence import Recurrence
from ..utils.telltime import getAwareDatetime, getLocalDatetime
from .vtimezone import create_timezone

# ------------------------------------------------------------------------------
class VComponentMixin:
    """Utilities for working with icalendar components"""
    def set(self, name, value, parameters=None, encode=1):
        if name in self:
            del self[name]
        self.add(name, value, parameters, encode)

class CalendarTypeError(TypeError):
    pass

class CalendarNotInitializedError(RuntimeError):
    pass

# ------------------------------------------------------------------------------
class ICalHandler:
    """Serve and load iCalendar files"""
    def serve(self, page, request, *args, **kwargs):
        try:
            vcal = VCalendar.fromPage(page, request)
        except CalendarTypeError:
            return None
        response = HttpResponse(vcal.to_ical(), content_type='text/calendar')
        response['Content-Disposition'] = \
            'attachment; filename={}.ics'.format(page.slug)
        return response

    def load(self, page, request, upload, **kwargs):
        vcal = VCalendar(page, **kwargs)
        vcal.load(request, upload.read(), getattr(upload, 'name', ""))

# ------------------------------------------------------------------------------
class VCalendar(Calendar, VComponentMixin):
    prodVersion = ".".join(__version__.split(".", 2)[:2])
    prodId = "-//linuxsoftware.nz//NONSGML Joyous v{}//EN".format(prodVersion)

    def __init__(self, page=None, utc2local=False):
        super().__init__(self)
        self.page = page
        self.utc2local = utc2local
        self.set('PRODID',  self.prodId)
        self.set('VERSION', "2.0")

    @classmethod
    def fromPage(cls, page, request):
        if isinstance(page, CalendarPage):
            return cls._fromCalendarPage(page, request)
        elif isinstance(page, EventBase):
            return cls._fromEventPage(page)
        else:
            raise CalendarTypeError("Unsupported input page")

    @classmethod
    def _fromCalendarPage(cls, page, request):
        vcal = cls(page)
        tzs = {}
        for event in page._getAllEvents(request):
            vevent = cls.factory.makeFromPage(event)
            vcal.add_component(vevent)
            for vchild in vevent.vchildren:
                vcal.add_component(vchild)
            if event.tz and event.tz is not pytz.utc:
                tzs.setdefault(event.tz, TimeZoneSpan()).add(vevent)
        for tz, vspan in tzs.items():
            vtz = vspan.createVTimeZone(tz)
            vcal.add_component(vtz)
        return vcal

    @classmethod
    def _fromEventPage(cls, event):
        vcal = cls(cls._findCalendarFor(event))
        vevent = cls.factory.makeFromPage(event)
        vcal.add_component(vevent)
        for vchild in vevent.vchildren:
            vcal.add_component(vchild)
        if event.tz and event.tz is not pytz.utc:
            vtz = TimeZoneSpan(vevent).createVTimeZone(event.tz)
            vcal.add_component(vtz)
        return vcal

    @classmethod
    def _findCalendarFor(cls, event):
        calendar = CalendarPage.objects.ancestor_of(event).first()
        if not calendar:
            site = event.get_site()
            if site:
                home = site.root_page
                calendar = CalendarPage.objects.descendant_of(home).first()
        if not calendar:
            calendar = CalendarPage.objects.first()
        return calendar

    def clear(self):
        super().clear()
        self.subcomponents.clear()

    def load(self, request, data, name=""):
        if self.page is None:
            raise CalendarNotInitializedError("No page set")

        # Typically, this information will consist of an iCalendar stream
        # with a single iCalendar object.  However, multiple iCalendar
        # objects can be sequentially grouped together in an iCalendar
        # stream.
        try:
            calStream = Calendar.from_ical(data, multiple=True)
        except ValueError as e:
            messages.error(request, "Could not parse iCalendar file "+name)
            #messages.debug(request, str(e))
            return

        self.clear()
        numSuccess = numFail = 0
        for cal in calStream:
            tz = timezone.get_current_timezone()
            zone = cal.get('X-WR-TIMEZONE', None)
            if zone:
                try:
                    tz = pytz.timezone(zone)
                except pytz.exceptions.UnknownTimeZoneError:
                    messages.warning(request, "Unknown time zone {}".format(zone))
            with timezone.override(tz):
                result = self._loadEvents(request, cal.walk(name="VEVENT"))
                numSuccess += result[0]
                numFail    += result[1]
        if numSuccess:
            messages.success(request, "{} iCal events loaded".format(numSuccess))
        if numFail:
            messages.error(request, "Could not load {} iCal events".format(numFail))

    def _loadEvents(self, request, vevents):
        numSuccess = numFail = 0
        vmap = {}
        for props in vevents:
            try:
                match = vmap.setdefault(str(props.get('UID')), VMatch())
                vevent = self.factory.makeFromProps(props, match.parent)
                if self.utc2local:
                    vevent._convertTZ()
            except CalendarTypeError as e:
                numFail += 1
                #messages.debug(request, str(e))
            else:
                self.add_component(vevent)
                match.add(vevent)

        for vmatch in vmap.values():
            vevent = vmatch.parent
            if vevent is not None:
                try:
                    event = self.page._getEventFromUid(request, vevent['UID'])
                except ObjectDoesNotExist:
                    numSuccess += self._createEventPage(request, vevent)
                else:
                    if event:
                        numSuccess += self._updateEventPage(request, vevent, event)
        return numSuccess, numFail

    def _updateEventPage(self, request, vevent, event):
        numUpdated = 0
        if vevent.modifiedDt > event.latest_revision_created_at:
            vevent.toPage(event)
            _saveRevision(request, event)
            numUpdated += 1

        vchildren  = vevent.vchildren[:]
        vchildren += [CancellationVEvent.fromExDate(vevent, exDate)
                      for exDate in vevent.exDates]
        for vchild in vchildren:
            try:
                exception = vchild.Page.objects.child_of(event)            \
                                  .get(except_date=vchild['RECURRENCE-ID'].date())
            except ObjectDoesNotExist:
                self._createExceptionPage(request, event, vchild)
            else:
                if exception.isAuthorized(request):
                    self._updateExceptionPage(request, vchild, exception)
        return numUpdated

    def _updateExceptionPage(self, request, vchild, exception):
        if vchild.modifiedDt > exception.latest_revision_created_at:
            vchild.toPage(exception)
            _saveRevision(request, exception)

    def _createEventPage(self, request, vevent):
        event = vevent.makePage(uid=vevent['UID'])
        _addPage(request, self.page, event)
        _saveRevision(request, event)
        numCreated = 1

        vchildren  = vevent.vchildren[:]
        vchildren += [CancellationVEvent.fromExDate(vevent, exDate)
                      for exDate in vevent.exDates]
        for vchild in vchildren:
            self._createExceptionPage(request, event, vchild)
        return numCreated

    def _createExceptionPage(self, request, event, vchild):
        exception = vchild.makePage(overrides=event)
        _addPage(request, event, exception)
        _saveRevision(request, exception)

# ------------------------------------------------------------------------------
def _addPage(request, parent, page):
    page.owner = request.user
    page.live  = bool(request.POST.get('action-publish'))
    parent.add_child(instance=page)

def _saveRevision(request, page):
    revision = page.save_revision(request.user,
                                  bool(request.POST.get('action-submit')))
    if bool(request.POST.get('action-publish')):
        revision.publish()

# ------------------------------------------------------------------------------
class vDt(vDDDTypes):
    """Smooths over some date-vs-datetime and aware-vs-naive differences"""
    def __init__(self, value=None, *, inclusive=False):
        self.dt = None
        if value is not None:
            if hasattr(value, 'dt'):
                value = value.dt
            if type(value) == dt.date and inclusive:
                value += dt.timedelta(days=1)
            super().__init__(value)

    def __bool__(self):
        return bool(self.dt)

    def __eq__(self, value):
        if hasattr(value, 'dt'):
            value = value.dt
        return self.dt == value

    def date(self, inclusive=False):
        if type(self.dt) == dt.datetime:
            return self.dt.date()
        elif type(self.dt) == dt.date:
            if inclusive:
                return self.dt - dt.timedelta(days=1)
            else:
                return self.dt

    def time(self):
        if type(self.dt) == dt.datetime:
            return self.dt.time()

    def datetime(self, timeDefault=dt.time.min):
        # TODO: Allow 'floating' times?
        tz = timezone.get_current_timezone()
        if type(self.dt) == dt.datetime:
            if timezone.is_aware(self.dt):
                return self.dt
            else:
                return timezone.make_aware(self.dt, tz)
        elif type(self.dt) == dt.date:
            return getAwareDatetime(self.dt, None, tz, timeDefault)

    def tzinfo(self):
        return getattr(self.dt, 'tzinfo', None)

    def zone(self):
        tzinfo = self.tzinfo()
        if hasattr(tzinfo, 'zone'):
            return tzinfo.zone
        elif hasattr(tzinfo, 'tzname'):
            return tzinfo.tzname(None)

    def timezone(self):
        tzinfo = self.tzinfo()
        if getattr(tzinfo, 'zone', None):
            try:
                # Return the timezone unbound from the datetime
                return pytz.timezone(tzinfo.zone)
            except pytz.exceptions.UnknownTimeZoneError as e:
                raise CalendarTypeError(str(e)) from e
        elif tzinfo is not None:
            return tzinfo
        else:
            return timezone.get_current_timezone()

class vSmart(vText):
    """Text property that automatically decodes encoded strings"""
    def __str__(self):
        retval = super().__str__()
        param = self.params.get('ENCODING', "").upper()
        if param == 'QUOTED-PRINTABLE':
            retval = quopri.decodestring(retval).decode(self.encoding, 'ignore')
        elif param == 'BASE64':
            retval = base64.b64decode(retval).decode(self.encoding, 'ignore')
        return retval

from icalendar.cal import types_factory
types_factory['date']      = vDt
types_factory['date-time'] = vDt
types_factory['text']      = vSmart

# ------------------------------------------------------------------------------
class TimeZoneSpan:
    """Combines common time zones"""
    class NotInitializedError(RuntimeError): pass

    def __init__(self, vevent=None):
        self.firstDt = None
        self.lastDt  = None
        if vevent is not None:
            self.add(vevent)

    def add(self, vevent):
        firstDt = vDt(vevent['DTSTART']).datetime()
        lastDt  = vDt(vevent['DTEND']).datetime(dt.time.max)

        rrule = vevent.get('RRULE', {})
        if rrule:
            # use until if it is available
            untilDt = vDt(rrule.get('UNTIL', [None])[0]).datetime(dt.time.max)
            if untilDt:
                lastDt = untilDt
            else:
                # pytz.timezones doesn't know any transition dates after 2038
                # either -- icalendar/src/icalendar/cal.py:526
                # using replace to keep the tzinfo
                lastDt = lastDt.replace(year=2038, month=12, day=31)

        if self.firstDt is None or firstDt < self.firstDt:
            self.firstDt = firstDt
        if self.lastDt is None or lastDt > self.lastDt:
            self.lastDt = lastDt

    def createVTimeZone(self, tz):
        if self.firstDt is None or self.lastDt is None:
            raise self.NotInitializedError()
        return create_timezone(tz, self.firstDt, self.lastDt)

# ------------------------------------------------------------------------------
class VMatch:
    """Matches recurring events with their exceptions"""
    class DuplicateError(RuntimeError): pass

    def __init__(self):
        self.parent  = None
        self.orphans = []

    def add(self, component):
        if isinstance(component, ExceptionVEvent):
            self._addChild(component)
        else:
            self._addParent(component)

    def _addChild(self, component):
        if self.parent:
            self.parent.vchildren.append(component)
        else:
            # I don't know of any iCalendar producer that lists exceptions
            # before the recurring event, but the RFC doesn't seem to say that
            # it can't happen either.
            self.orphans.append(component)

    def _addParent(self, component):
        if self.parent:
            raise self.DuplicateError("UID {}".format(component['UID']))
        self.parent = component
        for orphan in self.orphans:
            # reprocess orphans now their parent has been found
            vevent = VCalendar.factory.makeFromProps(orphan, self.parent)
            self.parent.vchildren.append(vevent)
        self.orphans.clear()

# ------------------------------------------------------------------------------
class VEventFactory:
    """Creates VEvent objects"""
    def makeFromProps(self, props, parent):
        if 'UID' not in props:
            raise CalendarTypeError("Missing UID")
        if 'DTSTAMP' not in props:
            raise CalendarTypeError("Missing DTSTAMP")
        dtstart = props.get('DTSTART')
        if not dtstart:
            raise CalendarTypeError("Missing DTSTART")

        dtend    = props.get('DTEND')
        duration = props.get('DURATION')
        if duration:
            if dtend:
                # dtend and duration must not occur in the same event properties
                raise CalendarTypeError("Both DURATION and DTEND set")
            else:
                dtend = props['DTEND'] = vDt(dtstart.dt + duration.dt)
        else:
            if not dtend:
                dtend = props['DTEND'] = vDt(dtstart, inclusive=True)
        if type(dtstart.dt) != type(dtend.dt):
            raise CalendarTypeError("DTSTART and DTEND types do not match")
        if dtstart.timezone() != dtend.timezone():
            # Yes it is valid, but Joyous does not support it
            raise CalendarTypeError("DTSTART.timezone != DTEND.timezone")

        numDays = (dtend.date(inclusive=True) - dtstart.date()).days + 1
        rrule = props.get('RRULE')
        if rrule is not None:
            if type(rrule) == list:
                # TODO support multiple RRULEs?
                raise CalendarTypeError("Multiple RRULEs")
            if numDays > 1:
                return MultidayRecurringVEvent.fromProps(props)
            else:
                return RecurringVEvent.fromProps(props)

        recurrenceId = props.get('RECURRENCE-ID')
        if recurrenceId is not None:
            if dtstart.timezone() != recurrenceId.timezone():
                # Also valid, but still Joyous does not support it
                raise CalendarTypeError("DTSTART.timezone != RECURRENCE-ID.timezone")

            if (parent and recurrenceId == dtstart and
                parent.numDays == numDays and
                parent['DTEND'].time() == dtend.time() and
                (parent['SUMMARY'] != props['SUMMARY'] or
                 parent['DESCRIPTION'] != props['DESCRIPTION'] != "")):
                return ExtraInfoVEvent.fromProps(props)
            else:
                if numDays > 1:
                    return RescheduleMultidayVEvent.fromProps(props)
                elif parent and parent.numDays > 1:
                    return RescheduleMultidayVEvent.fromProps(props)
                else:
                    return PostponementVEvent.fromProps(props)

        if (dtstart.date() != dtend.date(inclusive=True)):
            return MultidayVEvent.fromProps(props)
        else:
            return SimpleVEvent.fromProps(props)

    def makeFromPage(self, page):
        if isinstance(page, SimpleEventPage):
            return SimpleVEvent.fromPage(page)

        elif isinstance(page, MultidayEventPage):
            return MultidayVEvent.fromPage(page)

        elif isinstance(page, MultidayRecurringEventPage):
            return MultidayRecurringVEvent.fromPage(page)

        elif isinstance(page, RecurringEventPage):
            return RecurringVEvent.fromPage(page)

        elif isinstance(page, EventExceptionBase):
            return RecurringVEvent.fromPage(page.overrides)

        else:
            raise CalendarTypeError("Unsupported page type")

VCalendar.factory = VEventFactory()

# ------------------------------------------------------------------------------
class VEvent(Event, VComponentMixin):
    def __init__(self):
        super().__init__()
        self.vchildren = []

    @classmethod
    def fromPage(cls, page):
        vevent = cls()
        firstRevision = page.revisions.order_by("created_at").first()
        vevent.set('URL',           page.full_url)
        vevent.set('SUMMARY',       page.title)
        vevent.set('SEQUENCE',      page.revisions.count())
        vevent.set('DTSTAMP',       vDatetime(timezone.now()))
        vevent.set('CREATED',       vDatetime(firstRevision.created_at))
        vevent.set('LAST-MODIFIED', vDatetime(page.latest_revision_created_at))
        return vevent

    @classmethod
    def fromProps(cls, props):
        vevent = cls()
        vevent.update(props)
        return vevent

    def toPage(self, page):
        pass

    def makePage(self, **kwargs):
        if not hasattr(self, 'Page'):
            raise NotImplementedError()
        page = self.Page(**kwargs)
        self.toPage(page)
        return page

    def _convertTZ(self):
        """Will convert UTC datetimes to the current local timezone"""
        tz = timezone.get_current_timezone()
        dtstart = self['DTSTART']
        dtend   = self['DTEND']
        if dtstart.zone() == "UTC":
            dtstart.dt = dtstart.dt.astimezone(tz)
        if dtend.zone() == "UTC":
            dtend.dt = dtend.dt.astimezone(tz)

    @property
    def modifiedDt(self):
        prop = self.get('LAST-MODIFIED') or self.get('DTSTAMP')
        return prop.datetime()

    @property
    def exDates(self):
        exDates = self.get('EXDATE', [])
        if type(exDates) != list:
            exDates = [exDates]
        dtType = type(self['DTSTART'].dt)
        retval = {exDate for vddd in exDates for exDate in vddd.dts
                  if type(exDate.dt) == dtType}
        return list(retval)

    @property
    def numDays(self):
        dtstart = self['DTSTART']
        dtend   = self['DTEND']
        return (dtend.date(inclusive=True) - dtstart.date()).days + 1

# ------------------------------------------------------------------------------
class SimpleVEvent(VEvent):
    Page = SimpleEventPage

    @classmethod
    def fromPage(cls, page):
        vevent = super().fromPage(page)
        # FIXME support Anniversary date type events?
        dtstart = getAwareDatetime(page.date, page.time_from, page.tz, dt.time.min)
        dtend   = getAwareDatetime(page.date, page.time_to, page.tz, dt.time.max)
        vevent.set('UID',         page.uid)
        vevent.set('DTSTART',     vDatetime(dtstart))
        vevent.set('DTEND',       vDatetime(dtend))
        vevent.set('DESCRIPTION', page.details)
        vevent.set('LOCATION',    page.location)
        # TODO: Add CATEGORIES page.category
        # TODO: Add CLASS pages.get_view_restrictions() ? "RESTRICTED" : "PUBLIC"
        return vevent

    def toPage(self, page):
        super().toPage(page)
        assert page.uid == self.get('UID')
        page.title = str(self.get('SUMMARY', "")) or page.uid[:16]
        dtstart  = self['DTSTART']
        dtend    = self['DTEND']
        page.details    = str(self.get('DESCRIPTION', ""))
        page.location   = str(self.get('LOCATION', ""))
        page.date       = dtstart.date()
        page.time_from  = dtstart.time()
        page.time_to    = dtend.time()
        page.tz         = dtstart.timezone()

# ------------------------------------------------------------------------------
class MultidayVEvent(VEvent):
    Page = MultidayEventPage

    @classmethod
    def fromPage(cls, page):
        vevent = super().fromPage(page)
        # FIXME support Anniversary date type events?
        dtstart = getAwareDatetime(page.date_from, page.time_from, page.tz, dt.time.min)
        dtend   = getAwareDatetime(page.date_to, page.time_to, page.tz, dt.time.max)
        vevent.set('UID',         page.uid)
        vevent.set('DTSTART',     vDatetime(dtstart))
        vevent.set('DTEND',       vDatetime(dtend))
        vevent.set('DESCRIPTION', page.details)
        vevent.set('LOCATION',    page.location)
        return vevent

    def toPage(self, page):
        super().toPage(page)
        assert page.uid == self.get('UID')
        page.title = str(self.get('SUMMARY', "")) or page.uid[:16]
        dtstart  = self['DTSTART']
        dtend    = self['DTEND']
        page.details    = str(self.get('DESCRIPTION', ""))
        page.location   = str(self.get('LOCATION', ""))
        page.date_from  = dtstart.date()
        page.time_from  = dtstart.time()
        page.date_to    = dtend.date(inclusive=True)
        page.time_to    = dtend.time()
        page.tz         = dtstart.timezone()

# ------------------------------------------------------------------------------
class RecurringVEvent(VEvent):
    Page = RecurringEventPage

    @classmethod
    def fromPage(cls, page):
        vevent = super().fromPage(page)
        minDt   = pytz.utc.localize(dt.datetime.min)
        # FIXME support Anniversary date type events?
        dtstart = page._getMyFirstDatetimeFrom() or minDt
        dtend   = page._getMyFirstDatetimeTo()   or minDt
        vevent.set('UID',         page.uid)
        vevent.set('DTSTART',     vDatetime(dtstart))
        vevent.set('DTEND',       vDatetime(dtend))
        vevent.set('DESCRIPTION', page.details)
        vevent.set('LOCATION',    page.location)
        vevent.vchildren, exDates = cls.__getExceptions(page)
        if exDates:
            vevent.set('EXDATE', exDates)
        until = page.repeat.until
        if until:
            until = getAwareDatetime(until, dt.time.max, dtend.tzinfo)
            until = until.astimezone(pytz.utc)
        vevent.set('RRULE', vRecur.from_ical(page.repeat._getRrule(until)))
        return vevent

    @classmethod
    def __getExceptions(cls, page):
        vchildren = []
        exDates   = []
        for cancellation in CancellationPage.objects.live().child_of(page) \
                                            .select_related("postponementpage"):
            postponement = getattr(cancellation, "postponementpage", None)
            if postponement:
                vchildren.append(PostponementVEvent.fromPage(postponement))
            else:
                excludeDt = getAwareDatetime(cancellation.except_date,
                                             cancellation.time_from,
                                             cancellation.tz, dt.time.min)
                exDates.append(excludeDt)
                # NB any cancellation title or details are going to be lost
                # vchildren.append(CancellationVEvent.fromPage(cancellation))

            # if not postponement and cancellation.cancellation_title == "":
            #     excludeDt = getAwareDatetime(cancellation.except_date,
            #                                  cancellation.time_from,
            #                                  cancellation.tz, dt.time.min)
            #     exDates.append(excludeDt)
            # else:
            #     if cancellation.title != "":
            #         vchildren.append(CancellationVEvent.fromPage(cancellation))
            #     if postponement:
            #         vchildren.append(PostponementVEvent.fromPage(postponement))

        for info in ExtraInfoPage.objects.live().child_of(page):
            vchildren.append(ExtraInfoVEvent.fromPage(info))
        return vchildren, exDates

    def toPage(self, page):
        super().toPage(page)
        assert page.uid == self.get('UID')
        page.title = str(self.get('SUMMARY', "")) or page.uid[:16]
        dtstart  = self['DTSTART']
        dtend    = self['DTEND']
        rrule = self['RRULE']
        until = vDt(rrule.get('UNTIL', [None])[0])
        if until:
            untilDt = until.datetime().astimezone(dtstart.timezone())
            rrule['UNTIL'] = [untilDt.date()]
        page.details    = str(self.get('DESCRIPTION', ""))
        page.location   = str(self.get('LOCATION', ""))
        page.repeat     = Recurrence(rrule.to_ical().decode(),
                                     dtstart=dtstart.date())
        page.num_days   = (dtend.date() - dtstart.date()).days + 1
        page.time_from  = dtstart.time()
        page.time_to    = dtend.time()
        page.tz         = dtstart.timezone()

# ------------------------------------------------------------------------------
class MultidayRecurringVEvent(RecurringVEvent):
    Page = MultidayRecurringEventPage

# ------------------------------------------------------------------------------
class ExceptionVEvent(VEvent):
    def __init__(self):
        super().__init__()

    @classmethod
    def fromPage(cls, page):
        vevent = super().fromPage(page)
        daysDelta = dt.timedelta(days=page.num_days - 1)
        exceptDt = getAwareDatetime(page.except_date, page.overrides.time_from,
                                    page.tz, dt.time.min)
        dtstart  = getAwareDatetime(page.except_date, page.time_from,
                                    page.tz, dt.time.min)
        dtend    = getAwareDatetime(page.except_date + daysDelta, page.time_to,
                                    page.tz, dt.time.max)
        vevent.set('UID',           page.overrides.uid)
        vevent.set('RECURRENCE-ID', vDatetime(exceptDt))
        vevent.set('DTSTART',       vDatetime(dtstart))
        vevent.set('DTEND',         vDatetime(dtend))
        vevent.set('LOCATION',      page.overrides.location)
        return vevent

    def toPage(self, page):
        super().toPage(page)
        assert page.overrides.uid == self.get('UID')
        page.except_date = self['RECURRENCE-ID'].date()

# ------------------------------------------------------------------------------
class ExtraInfoVEvent(ExceptionVEvent):
    Page = ExtraInfoPage

    @classmethod
    def fromPage(cls, page):
        vevent = super().fromPage(page)
        vevent.set('SUMMARY',       page.extra_title or page.overrides.title)
        vevent.set('DESCRIPTION',   page.extra_information)
        return vevent

    def toPage(self, page):
        super().toPage(page)
        page.extra_title       = str(self.get('SUMMARY', ""))
        page.extra_information = str(self.get('DESCRIPTION', ""))

# ------------------------------------------------------------------------------
class CancellationVEvent(ExceptionVEvent):
    Page = CancellationPage

    def property_items(self, recursive=True, sorted=True):
        # Cancellations are represented by EXDATE in VEVENT not as their own
        # VEVENT instance.  Yes, this means the cancellation_title and
        # cancellation_details are lost.
        return []

    @classmethod
    def fromExDate(cls, vparent, exDate):
        vevent = cls()
        vevent.set('DTSTAMP',       vparent['DTSTAMP'])
        vevent.set('UID',           vparent['UID'])
        vevent.set('RECURRENCE-ID', vDt(exDate))
        vparent.vchildren.append(vevent)
        return vevent

    @classmethod
    def fromPage(cls, page):
        vevent = super().fromPage(page)
        vevent.set('SUMMARY',       page.cancellation_title)
        vevent.set('DESCRIPTION',   page.cancellation_details)
        return vevent

    def toPage(self, page):
        super().toPage(page)
        page.cancellation_title   = str(self.get('SUMMARY', ""))
        page.cancellation_details = str(self.get('DESCRIPTION', ""))

# ------------------------------------------------------------------------------
class PostponementVEvent(ExceptionVEvent):
    Page = PostponementPage

    @classmethod
    def fromPage(cls, page):
        vevent = super().fromPage(page)
        daysDelta = dt.timedelta(days=page.num_days - 1)
        dtstart = getAwareDatetime(page.date, page.time_from,
                                   page.tz, dt.time.min)
        dtend   = getAwareDatetime(page.date + daysDelta, page.time_to,
                                   page.tz, dt.time.max)
        vevent.set('SUMMARY',     page.postponement_title)
        vevent.set('DTSTART',     vDatetime(dtstart))
        vevent.set('DTEND',       vDatetime(dtend))
        vevent.set('DESCRIPTION', page.details)
        vevent.set('LOCATION',    page.location)
        return vevent

    def toPage(self, page):
        super().toPage(page)
        dtstart  = self['DTSTART']
        dtend    = self['DTEND']
        page.postponement_title = str(self.get('SUMMARY', ""))
        page.details            = str(self.get('DESCRIPTION', ""))
        page.location           = str(self.get('LOCATION', ""))
        page.date               = dtstart.date()
        page.time_from          = dtstart.time()
        page.time_to            = dtend.time()

    def makePage(self, **kwargs):
        return super().makePage(**kwargs)

# ------------------------------------------------------------------------------
class RescheduleMultidayVEvent(PostponementVEvent):
    Page = RescheduleMultidayEventPage

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Many things utilities
# ------------------------------------------------------------------------------
from django.utils.translation import get_language
from django.utils.translation import to_locale
from django.utils.translation import gettext as _, gettext_noop
from num2words import num2words

# ------------------------------------------------------------------------------
def _n2w(n, to):
    try:
        return num2words(n, lang=to_locale(get_language()), to=to)
    except NotImplementedError:
        # fall back to gettext for these words
        gettext_noop("first")
        gettext_noop("second")
        gettext_noop("third")
        gettext_noop("fourth")
        gettext_noop("fifth")
        return _(num2words(n, lang="en", to=to))

# ------------------------------------------------------------------------------
def toOrdinal(n):
    """
    Returns the ordinal name of a number
    e.g. 'first'
    """
    if n == -1:
        retval = _("last")
    elif n == -2:
        retval = _("penultimate")
    elif 1 <= n <= 5:
        retval = _n2w(n, to="ordinal")
    else:
        retval = _n2w(n, to="ordinal_num")
    return retval

# ------------------------------------------------------------------------------
def toTheOrdinal(n, inTitleCase=True):
    """
    Returns the definite article with the ordinal name of a number
    e.g. 'the second'
    Becomes important for languages with multiple definite articles (e.g. French)
    """
    if n == -1:
        retval = _("the last")
    elif n == -2:
        retval = _("the penultimate")
    elif n == 1:
        retval = _("the first")
    elif n == 2:
        retval = _("the second")
    elif n == 3:
        retval = _("the third")
    elif n == 4:
        retval = _("the fourth")
    elif n == 5:
        retval = _("the fifth")
    else:
        # TODO: support other definite articles depending on gender, etc.
        retval = _("the")
        if inTitleCase:
            retval = retval.title()
        retval += " "+_n2w(n, to="ordinal_num")
        return retval
    if inTitleCase:
        retval = retval.title()
    return retval

# ------------------------------------------------------------------------------
def hrJoin(items):
    """
    Joins items together in a human readable string
    e.g. 'wind, ice and fire'
    """
    conjuction = " "+_("and")+" "
    if len(items) <= 2:
        return conjuction.join(items)
    else:
        return ", ".join(items[:-1]) + conjuction + items[-1]

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Useful Mixins
# ------------------------------------------------------------------------------
from django.contrib.contenttypes.models import ContentType


# ------------------------------------------------------------------------------
class ProxyPageMixin:
    """
    Adding this mixin allows inheritence without needing a new table.  The
    proxy model has its own content type which allows it to be selected as a
    separate page type in the Wagtail admin interface.  No change is made to
    the manager, but peers() will return a queryset of others of the same type.
    See also https://github.com/wagtail/wagtail/pull/1736
    """
    class Meta:
        proxy = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._initContentType()

    def _initContentType(self):
        if getattr(self, 'id', object()) is None:
            self.content_type = self._getContentType()

    @classmethod
    def peers(cls):
        """Return others of the same type"""
        return cls.objects.filter(content_type=cls._getContentType())

    @classmethod
    def _getContentType(cls):
        return ContentType.objects.get_for_model(cls, for_concrete_model=False)

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# Generated by Django 2.1.4 on 2018-12-31 05:27

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import ls.joyous.utils.mixins


class Migration(migrations.Migration):

    dependencies = [
        ('joyous', '0011_auto_20180921_0700'),
    ]

    operations = [
        migrations.CreateModel(
            name='MultidayRecurringEventPage',
            fields=[
                ('recurringeventpage_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='joyous.RecurringEventPage')),
            ],
            options={
                'verbose_name': 'Multiday Recurring Event Page',
            },
            bases=(ls.joyous.utils.mixins.ProxyPageMixin, 'joyous.recurringeventpage'),
        ),
        migrations.AddField(
            model_name='recurringeventpage',
            name='num_days',
            field=models.IntegerField(default=1, validators=[django.core.validators.MinValueValidator(1), django.core.validators.MaxValueValidator(99)], verbose_name='Number of days'),
        ),
    ]



# Generated by Django 2.1.5 on 2019-04-08 18:45

import django.core.validators
from django.db import migrations, models
import ls.joyous.utils.mixins


class Migration(migrations.Migration):

    dependencies = [
        ('wagtailcore', '0040_page_draft_title'),
        ('wagtailredirects', '0006_redirect_increase_max_length'),
        ('wagtailforms', '0003_capitalizeverbose'),
        ('joyous', '0014_auto_20190328_0652'),
    ]

    operations = [
        migrations.CreateModel(
            name='RescheduleMultidayEventPage',
            fields=[
            ],
            options={
                'indexes': [],
                'verbose_name_plural': 'postponements',
                'proxy': True,
                'verbose_name': 'postponement',
            },
            bases=(ls.joyous.utils.mixins.ProxyPageMixin, 'joyous.postponementpage'),
        ),
        migrations.RemoveField(
            model_name='postponementpage',
            name='group_page',
        ),
        migrations.RemoveField(
            model_name='postponementpage',
            name='uid',
        ),
        migrations.AddField(
            model_name='postponementpage',
            name='num_days',
            field=models.IntegerField(default=1, validators=[django.core.validators.MinValueValidator(1), django.core.validators.MaxValueValidator(99)], verbose_name='number of days'),
        ),
        migrations.DeleteModel(
            name='MultidayRecurringEventPage',
        ),
        migrations.CreateModel(
            name='MultidayRecurringEventPage',
            fields=[
            ],
            options={
                'indexes': [],
                'verbose_name_plural': 'multiday recurring event pages',
                'proxy': True,
                'verbose_name': 'multiday recurring event page',
            },
            bases=(ls.joyous.utils.mixins.ProxyPageMixin, 'joyous.recurringeventpage'),
        ),
    ]



# ------------------------------------------------------------------------------
# Names for i18n
# ------------------------------------------------------------------------------
import sys
from django.utils.translation import gettext_lazy as _
from django.utils import dates

# ------------------------------------------------------------------------------
class _Names(tuple):
    """
    Holds internationalized names and only translates them at the last possible
    moment, when they are accessed.
    """
    def __getitem__(self, i):
        return str(super().__getitem__(i))

# ------------------------------------------------------------------------------
#: Names of days of the week, from Monday to Sunday
MONDAY_TO_SUNDAY = WEEKDAY_NAMES = _Names(dates.WEEKDAYS[k] for k in range(7))

#: Names of days of the week, from Sunday to Saturday
SUNDAY_TO_SATURDAY = _Names(dates.WEEKDAYS[k%7] for k in range(6,13))

#: Abbreviations of days of the week, from Mon to Sun
MON_TO_SUN = WEEKDAY_ABBRS = _Names(dates.WEEKDAYS_ABBR[k] for k in range(7))

#: Abbreviations of days of the week, from Sun to Sat
SUN_TO_SAT = _Names(dates.WEEKDAYS_ABBR[k%7] for k in range(6,13))

# ------------------------------------------------------------------------------

#: The names of days of the week in plural, Mondays to Sundays
WEEKDAY_NAMES_PLURAL = _Names((_("Mondays"),
                               _("Tuesdays"),
                               _("Wednesdays"),
                               _("Thursdays"),
                               _("Fridays"),
                               _("Saturdays"),
                               _("Sundays")))

# ------------------------------------------------------------------------------
#: Names of the months, with January at index 1
MONTH_NAMES = _Names(dates.MONTHS.get(k,"") for k in range(13))

#: Names of the months, December, January, ...December, January
WRAPPED_MONTH_NAMES = _Names(dates.MONTHS[k%12+1] for k in range(11,25))

#: Abbreviations of the months, with Jan at index 1
MONTH_ABBRS = _Names(dates.MONTHS_3.get(k, "") for k in range(13))

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ----------------------
#  ls.joyous
# ----------------------

import sys
import os
import io
import subprocess
from pathlib import Path
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

# allow setup.py to be run from any path
here = Path(__file__).resolve().parent
os.chdir(str(here))

with io.open("README.rst", encoding="utf-8") as readme:
    README = readme.read()

class RunTests(TestCommand):
    def run(self):
        errno = subprocess.call([sys.executable, 'runtests.py'])
        raise SystemExit(errno)

setup(name="ls.joyous",
      use_scm_version={
          'write_to':  "ls/joyous/_version.py",
      },
      description="A calendar application for Wagtail.",
      long_description=README,
      keywords=["calendar", "events", "wagtail", "groupware"],
      classifiers=["Development Status :: 4 - Beta",
                   "Framework :: Django",
                   "Framework :: Wagtail",
                   "Framework :: Wagtail :: 2",
                   "License :: OSI Approved :: BSD License",
                   "Operating System :: OS Independent",
                   "Programming Language :: Python",
                   "Programming Language :: Python :: 3",
                   "Topic :: Office/Business :: Groupware",
                   "Topic :: Office/Business :: Scheduling",
                   "Topic :: Software Development :: Libraries :: Python Modules"
                  ],
      platforms="any",
      author="David Moore",
      author_email="david@linuxsoftware.co.nz",
      url="https://github.com/linuxsoftware/ls.joyous",
      license="BSD",
      packages=find_packages(where=".", exclude=["ls.joyous.tests"]),
      package_data={'ls.joyous': ["templates/joyous/*.html",
                                  "templates/joyous/*/*.html",
                                  "static/joyous/css/*.css",
                                  "static/joyous/css/vendor/*.css",
                                  "static/joyous/img/*.png",
                                  "static/joyous/img/*.jpg",
                                  "static/joyous/js/*.js",
                                  "static/joyous/js/vendor/*.js",
                                  "locale/*/LC_MESSAGES/django.po",
                                  "locale/*/LC_MESSAGES/django.mo"
                                 ],
                   },
      setup_requires=["setuptools_scm"],
      install_requires=["django-timezone-field",
                        "holidays",
                        "icalendar",
                        "num2words",
                        "python-dateutil",
                       ],
      tests_require=["coverage",
                     "django-beautifulsoup-test",
                     "freezegun"],
      test_suite="ls.joyous.tests",
      cmdclass={'test': RunTests},
      zip_safe=False,
     )



# ------------------------------------------------------------------------------
# Test Postponement Page
# ------------------------------------------------------------------------------
import sys
import pytz
from freezegun import freeze_time
import datetime as dt
from django.test import RequestFactory, TestCase
from django.contrib.auth.models import User
from django.utils import timezone
from wagtail.core.models import Page
from ls.joyous.models import (GeneralCalendarPage, RecurringEventPage,
        MultidayRecurringEventPage, CancellationPage, PostponementPage,
        RescheduleMultidayEventPage)
from ls.joyous.utils.recurrence import Recurrence, WEEKLY, MO, TU, WE, FR

# ------------------------------------------------------------------------------
class Test(TestCase):
    def setUp(self):
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('j', 'j@joy.test', 's3(r3t')
        self.request = RequestFactory().get("/test")
        self.request.user = self.user
        self.request.session = {}
        self.calendar = GeneralCalendarPage(owner = self.user,
                                            slug  = "events",
                                            title = "Events")
        self.home.add_child(instance=self.calendar)
        self.calendar.save_revision().publish()
        self.event = RecurringEventPage(slug   = "test-meeting",
                                        title  = "Test Meeting",
                                        repeat = Recurrence(dtstart=dt.date(1990,1,1),
                                                            freq=WEEKLY,
                                                            byweekday=[MO,WE,FR]),
                                        time_from = dt.time(13,30),
                                        time_to   = dt.time(16))
        self.calendar.add_child(instance=self.event)
        self.postponement = PostponementPage(owner = self.user,
                                             overrides = self.event,
                                             except_date = dt.date(1990,10,10),
                                             cancellation_title   = "Meeting Postponed",
                                             cancellation_details =
                                                 "The meeting has been postponed until tomorrow",
                                             postponement_title   = "A Meeting",
                                             date      = dt.date(1990,10,11),
                                             time_from = dt.time(13),
                                             time_to   = dt.time(16,30),
                                             details   = "Yes a test meeting on a Thursday")
        self.event.add_child(instance=self.postponement)
        self.postponement.save_revision().publish()

    def testGetEventsByDay(self):
        events = RecurringEventPage.events.byDay(dt.date(1990,10,1),
                                                 dt.date(1990,10,31))
        self.assertEqual(len(events), 31)
        evod = events[9]
        self.assertEqual(evod.date, dt.date(1990,10,10))
        self.assertEqual(len(evod.days_events), 1)
        self.assertEqual(len(evod.continuing_events), 0)
        title, page = evod.days_events[0]
        self.assertEqual(title, "Meeting Postponed")
        self.assertIs(type(page), CancellationPage)
        self.assertIs(type(page.postponementpage), PostponementPage)

        events = PostponementPage.events.byDay(dt.date(1990,10,1),
                                               dt.date(1990,10,31))
        self.assertEqual(len(events), 31)
        evod = events[10]
        self.assertEqual(evod.date, dt.date(1990,10,11))
        self.assertEqual(len(evod.days_events), 1)
        self.assertEqual(len(evod.continuing_events), 0)
        title, page = evod.days_events[0]
        self.assertEqual(title, "A Meeting")
        self.assertIs(type(page), PostponementPage)

    def testStatus(self):
        self.assertEqual(self.postponement.status, "finished")
        self.assertEqual(self.postponement.status_text, "This event has finished.")
        now = timezone.localtime()
        myday = now.date() + dt.timedelta(1)
        friday = myday + dt.timedelta(days=(4-myday.weekday())%7)
        futureEvent = PostponementPage(owner = self.user,
                                       overrides = self.event,
                                       except_date = friday,
                                       cancellation_title   = "",
                                       cancellation_details = "",
                                       postponement_title   = "Tuesday Meeting",
                                       date      = friday + dt.timedelta(days=4),
                                       time_from = dt.time(13,30),
                                       time_to   = dt.time(16),
                                       details   = "The meeting postponed from last Friday")
        self.event.add_child(instance=futureEvent)
        self.assertIsNone(futureEvent.status)
        self.assertEqual(futureEvent.status_text, "")

    def testWhen(self):
        self.assertEqual(self.postponement.when, "Thursday 11th of October 1990 at 1pm to 4:30pm")

    @freeze_time("2017-05-01")
    def testAt(self):
        self.assertEqual(self.postponement.at.strip(), "1pm")
        nextDate = self.event.next_date
        newDate  = nextDate + dt.timedelta(1)
        reschedule = PostponementPage(owner = self.user,
                                      overrides = self.event,
                                      except_date = nextDate,
                                      cancellation_title   = "",
                                      cancellation_details = "",
                                      postponement_title   = "Early Meeting",
                                      date      = newDate,
                                      time_from = dt.time(8,30),
                                      time_to   = dt.time(11),
                                      details   = "The meeting will be held early tomorrow")
        self.event.add_child(instance=reschedule)
        nextOn = self.event._nextOn(self.request)
        url = "/events/test-meeting/{}-postponement/".format(nextDate)
        self.assertEqual(nextOn[:76], '<a class="inline-link" href="{}">'.format(url))
        self.assertEqual(nextOn[-4:], '</a>')
        parts = nextOn[76:-4].split()
        self.assertEqual(len(parts), 6)
        self.assertEqual(parts[0], "{:%A}".format(newDate))
        self.assertEqual(int(parts[1][:-2]), newDate.day)
        self.assertIn(parts[1][-2:], ["st", "nd", "rd", "th"])
        self.assertEqual(parts[2], "of")
        self.assertEqual(parts[3], "{:%B}".format(newDate))
        self.assertEqual(parts[4], "at")
        self.assertEqual(parts[5], "8:30am")

# ------------------------------------------------------------------------------
class TestMultiday(TestCase):
    def setUp(self):
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('j', 'j@joy.test', 's3(r3t')
        self.request = RequestFactory().get("/test")
        self.request.user = self.user
        self.request.session = {}
        self.calendar = GeneralCalendarPage(owner = self.user,
                                            slug  = "events",
                                            title = "Events")
        self.home.add_child(instance=self.calendar)
        self.calendar.save_revision().publish()
        self.event = MultidayRecurringEventPage(slug = "test-session",
                                        title = "Test Session",
                                        repeat = Recurrence(dtstart=dt.date(1990,1,2),
                                                            freq=WEEKLY,
                                                            byweekday=[TU],
                                                            until=dt.date(1990,3,29)),
                                        num_days  = 3,
                                        time_from = dt.time(10),
                                        time_to   = dt.time(16,30))
        self.calendar.add_child(instance=self.event)
        self.postponement = RescheduleMultidayEventPage(owner = self.user,
                                             overrides = self.event,
                                             except_date = dt.date(1990,1,9),
                                             postponement_title   = "Delayed Start Session",
                                             date      = dt.date(1990,1,9),
                                             num_days  = 3,
                                             time_from = dt.time(13),
                                             time_to   = dt.time(19,30))
        self.event.add_child(instance=self.postponement)
        self.postponement.save_revision().publish()

    def testWhen(self):
        self.assertEqual(self.postponement.when,
                         "Tuesday 9th of January 1990 for 3 days "
                         "starting at 1pm finishing at 7:30pm")

# ------------------------------------------------------------------------------
class TestTZ(TestCase):
    def setUp(self):
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('j', 'j@joy.test', 's3(r3t')
        self.calendar = GeneralCalendarPage(owner = self.user,
                                            slug  = "events",
                                            title = "Events")
        self.home.add_child(instance=self.calendar)
        self.calendar.save_revision().publish()
        self.event = RecurringEventPage(slug   = "test-meeting",
                                        title  = "Test Meeting",
                                        repeat = Recurrence(dtstart=dt.date(1990,1,1),
                                                            freq=WEEKLY,
                                                            byweekday=[MO,WE,FR]),
                                        time_from = dt.time(13,30),
                                        time_to   = dt.time(16),
                                        tz = pytz.timezone("US/Eastern"))
        self.calendar.add_child(instance=self.event)
        self.postponement = PostponementPage(owner = self.user,
                                             overrides = self.event,
                                             postponement_title = "Delayed Meeting",
                                             except_date = dt.date(1990,10,10),
                                             date      = dt.date(1990,10,11),
                                             time_from = dt.time(13),
                                             time_to   = dt.time(16,30))
        self.event.add_child(instance=self.postponement)
        self.postponement.save_revision().publish()

    @timezone.override("Pacific/Auckland")
    def testLocalTitle(self):
        self.assertEqual(self.postponement.title,
                         "Postponement for Wednesday 10th of October 1990")
        self.assertEqual(self.postponement.local_title,
                         "Postponement for Thursday 11th of October 1990")

    @timezone.override("Asia/Colombo")
    def testGetEventsByDay(self):
        events = PostponementPage.events.byDay(dt.date(1990,10,1),
                                               dt.date(1990,10,31))
        self.assertEqual(len(events), 31)
        evod0 = events[10]
        self.assertEqual(evod0.date, dt.date(1990,10,11))
        self.assertEqual(len(evod0.days_events), 1)
        self.assertEqual(len(evod0.continuing_events), 0)
        title, page = evod0.days_events[0]
        self.assertEqual(title, "Delayed Meeting")
        self.assertIs(type(page), PostponementPage)
        evod1 = events[11]
        self.assertEqual(evod1.date, dt.date(1990,10,12))
        self.assertEqual(len(evod1.days_events), 0)
        self.assertEqual(len(evod1.continuing_events), 1)
        title, page = evod1.continuing_events[0]
        self.assertEqual(title, "Delayed Meeting")
        self.assertIs(type(page), PostponementPage)

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Test Get Event Functions
# ------------------------------------------------------------------------------
import sys
import datetime as dt
import pytz
import calendar
from django.test import RequestFactory, TestCase
from django.contrib.auth.models import User, AnonymousUser, Group
from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from django.utils import timezone
from wagtail.core.models import Site, Page, PageViewRestriction
from ls.joyous.utils.recurrence import Recurrence
from ls.joyous.utils.recurrence import WEEKLY, MONTHLY, MO, TU, WE, FR, SU
from ls.joyous.models.calendar import GeneralCalendarPage
from ls.joyous.models.events import (SimpleEventPage, MultidayEventPage,
        RecurringEventPage, PostponementPage, ExtraInfoPage)
from ls.joyous.models.events import (getAllEventsByDay, getAllEventsByWeek,
        getAllUpcomingEvents, getAllPastEvents, getGroupUpcomingEvents,
        getEventFromUid)
from ls.joyous.models.groups import get_group_model
from .testutils import datetimetz

GroupPage = get_group_model()

# ------------------------------------------------------------------------------
class Test(TestCase):
    def setUp(self):
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('i', 'i@foo.test', 's3cr3t')
        self.request = RequestFactory().get("/test")
        self.request.user = self.user
        self.request.session = {}
        self.calendar = GeneralCalendarPage(owner = self.user,
                                            slug  = "events",
                                            title = "Events")
        self.home.add_child(instance=self.calendar)
        self.group = GroupPage(slug = "initech", title = "Initech Corporation")
        self.home.add_child(instance=self.group)

        self.show = SimpleEventPage(owner = self.user,
                                    slug   = "pet-show",
                                    title  = "Pet Show",
                                    date      = dt.date(2013,1,5),
                                    time_from = dt.time(11),
                                    time_to   = dt.time(17,30),
                                    uid       = "29daefed-fed1-4e47-9408-43ec9b06a06d")
        self.calendar.add_child(instance=self.show)

        GROUPS = PageViewRestriction.GROUPS
        self.friends = Group.objects.create(name = "Friends")
        self.rendezvous = SimpleEventPage(owner = self.user,
                                          slug   = "rendezvous",
                                          title  = "Private Rendezvous",
                                          date      = dt.date(2013,1,10),
                                          uid       = "80af64e7-84e6-40d9-8b4f-7edf92aab9f7")
        self.calendar.add_child(instance=self.rendezvous)
        self.rendezvous.save_revision().publish()
        restriction = PageViewRestriction.objects.create(restriction_type = GROUPS,
                                                         page = self.rendezvous)
        restriction.groups.set([self.friends])
        restriction.save()

        self.party = MultidayEventPage(owner = self.user,
                                       slug  = "allnighter",
                                       title = "All Night",
                                       date_from = dt.date(2012,12,31),
                                       date_to   = dt.date(2013,1,1),
                                       time_from = dt.time(23),
                                       time_to   = dt.time(3),
                                       uid       = "initiative+technology")
        self.calendar.add_child(instance=self.party)

        self.standup = RecurringEventPage(slug   = "test-meeting",
                                          title  = "Test Meeting",
                                          repeat = Recurrence(dtstart=dt.date(2013,1,1),
                                                              until=dt.date(2013,5,31),
                                                              freq=WEEKLY,
                                                              byweekday=[MO,WE,FR]),
                                          time_from = dt.time(13,30),
                                          time_to   = dt.time(16),
                                          uid       = "initiative+technology")
        self.group.add_child(instance=self.standup)

        self.postponement = PostponementPage(owner = self.user,
                                             slug  = "2013-01-09-postponement",
                                             title = "Postponement for Wednesday 16th of October",
                                             overrides = self.standup,
                                             except_date = dt.date(2013,1,16),
                                             cancellation_title   = "Meeting Postponed",
                                             cancellation_details =
                                                 "The meeting has been postponed until tomorrow",
                                             postponement_title   = "A Meeting",
                                             date      = dt.date(2013,1,17),
                                             time_from = dt.time(13),
                                             time_to   = dt.time(16,30),
                                             details   = "Yes a test meeting on a Thursday")
        self.standup.add_child(instance=self.postponement)

    def testGetAllEventsByDay(self):
        events = getAllEventsByDay(self.request, dt.date(2013,1,1), dt.date(2013,1,31))
        self.assertEqual(len(events), 31)
        evod1 = events[0]
        self.assertEqual(evod1.date, dt.date(2013,1,1))
        self.assertEqual(len(evod1.all_events), 1)
        self.assertEqual(len(evod1.days_events), 0)
        self.assertEqual(len(evod1.continuing_events), 1)
        evod10 = events[9]
        self.assertEqual(evod10.date, dt.date(2013,1,10))
        self.assertEqual(len(evod10.all_events), 0)

    def testAuthGetAllEventsByDay(self):
        self.request.user.groups.set([self.friends])
        events = getAllEventsByDay(self.request, dt.date(2013,1,1), dt.date(2013,1,31))
        self.assertEqual(len(events), 31)
        evod10 = events[9]
        self.assertEqual(evod10.date, dt.date(2013,1,10))
        self.assertEqual(len(evod10.all_events), 1)
        self.assertEqual(len(evod10.days_events), 1)
        event = evod10.days_events[0]
        self.assertEqual(event.title, "Private Rendezvous")

    def testGetAllEventsByWeek(self):
        weeks = getAllEventsByWeek(self.request, 2013, 1)
        self.assertEqual(len(weeks), 5)
        self.assertIsNone(weeks[0][0])
        self.assertIsNone(weeks[0][1])
        self.assertIsNone(weeks[4][5])
        self.assertIsNone(weeks[4][6])
        evod = weeks[2][4]
        self.assertEqual(evod.date, dt.date(2013,1,17))
        self.assertEqual(len(evod.days_events), 1)
        self.assertEqual(len(evod.continuing_events), 0)

    def testGetAllUpcomingEvents(self):
        today = timezone.localdate()
        futureEvent = MultidayEventPage(owner = self.user,
                                        slug  = "tomorrow",
                                        title = "Tomorrow's Event",
                                        date_from = today + dt.timedelta(days=1),
                                        date_to   = today + dt.timedelta(days=3),
                                        time_from = dt.time(17),
                                        time_to   = dt.time(10,30))
        self.calendar.add_child(instance=futureEvent)
        events = getAllUpcomingEvents(self.request, home=self.home)
        self.assertEqual(len(events), 1)
        title, event = events[0]
        self.assertEqual(title, "Tomorrow's Event")
        self.assertEqual(event.slug, "tomorrow")
        events0 = getAllUpcomingEvents(self.request)
        self.assertEqual(len(events0), 1)

    def testGetAllPastEvents(self):
        events = getAllPastEvents(self.request)
        self.assertEqual(len(events), 4)
        self.assertEqual(events[0].title, "Test Meeting")
        self.assertEqual(events[1].title, "A Meeting")
        self.assertEqual(events[2].title, "Pet Show")
        self.assertEqual(events[3].title, "All Night")

    def testGetGroupUpcomingEvents(self):
        meeting = RecurringEventPage(owner = self.user,
                                     slug  = "plan-plan",
                                     title = "Planning to Plan",
                                     repeat    = Recurrence(dtstart=dt.date(2018,5,1),
                                                            freq=WEEKLY,
                                                            byweekday=[TU]),
                                     time_from = dt.time(18,30),
                                     time_to   = dt.time(20),
                                     group_page = self.group)
        self.calendar.add_child(instance=meeting)
        memo = ExtraInfoPage(owner = self.user,
                             slug  = "plan-plan-extra-info",
                             title = "Extra Information Planning to Plan",
                             overrides = meeting,
                             except_date = meeting.next_date,
                             extra_title = "Gap Analysis",
                             extra_information = "Analyse your gaps")
        meeting.add_child(instance=memo)
        events = getGroupUpcomingEvents(self.request, self.group)
        self.assertEqual(len(events), 2)
        self.assertEqual(events[0].title, "Gap Analysis")
        self.assertEqual(events[1].title, "Planning to Plan")

    def testGetGroupUpcomingEventsDupGroup(self):
        meeting = RecurringEventPage(owner = self.user,
                                     slug  = "plan-plan",
                                     title = "Planning to Plan",
                                     repeat    = Recurrence(dtstart=dt.date(2018,5,2),
                                                            freq=WEEKLY,
                                                            byweekday=[WE]),
                                     time_from = dt.time(18,30),
                                     time_to   = dt.time(20),
                                     group_page = self.group)
        self.group.add_child(instance=meeting)
        events = getGroupUpcomingEvents(self.request, self.group)
        self.assertEqual(len(events), 1)
        self.assertEqual(events[0].title, "Planning to Plan")
        self.assertEqual(events[0].page.group, self.group)

    def testGetGroupUpcomingEvents2Groups(self):
        rival = GroupPage(slug = "initrode", title = "Initrode Corporation")
        self.home.add_child(instance=rival)
        meeting = RecurringEventPage(owner = self.user,
                                     slug  = "plan-plan",
                                     title = "Planning to Plan",
                                     repeat    = Recurrence(dtstart=dt.date(2018,5,2),
                                                            freq=WEEKLY,
                                                            byweekday=[WE]),
                                     time_from = dt.time(18,30),
                                     time_to   = dt.time(20),
                                     group_page = rival)
        self.group.add_child(instance=meeting)
        events = getGroupUpcomingEvents(self.request, rival)
        self.assertEqual(len(events), 1)
        self.assertEqual(events[0].title, "Planning to Plan")
        # being a child of self.group trumps having group_page set
        self.assertEqual(events[0].page.group, self.group)
        events = getGroupUpcomingEvents(self.request, self.group)
        self.assertEqual(len(events), 1)
        self.assertEqual(events[0].title, "Planning to Plan")
        self.assertEqual(events[0].page.group, self.group)

    def testGetEventFromUid(self):
        event = getEventFromUid(self.request, "29daefed-fed1-4e47-9408-43ec9b06a06d")
        self.assertEqual(event.title, "Pet Show")

    def testMultiGetEventFromUid(self):
        with self.assertRaises(MultipleObjectsReturned):
            getEventFromUid(self.request, "initiative+technology")

    def testMissingGetEventFromUid(self):
        with self.assertRaises(ObjectDoesNotExist):
            getEventFromUid(self.request, "d12971fb-e694-4a04-aba2-fb1a4a7166b9")

    def testAuthGetEventFromUid(self):
        event = getEventFromUid(self.request, "80af64e7-84e6-40d9-8b4f-7edf92aab9f7")
        self.assertIsNone(event)
        self.request.user.groups.set([self.friends])
        event = getEventFromUid(self.request, "80af64e7-84e6-40d9-8b4f-7edf92aab9f7")
        self.assertIsNotNone(event.title)
        self.assertEqual(event.title, "Private Rendezvous")

# ------------------------------------------------------------------------------
class TestTZ(TestCase):
    def setUp(self):
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('i', 'i@foo.test', 's3cr3t')
        self.request = RequestFactory().get("/test")
        self.request.user = self.user
        self.request.session = {}
        self.calendar = GeneralCalendarPage(owner = self.user,
                                            slug  = "events",
                                            title = "Events")
        self.home.add_child(instance=self.calendar)

        self.night = RecurringEventPage(slug   = "pacnight",
                                        title  = "Pacific Night",
                                        repeat = Recurrence(dtstart=dt.date(2018,12,1),
                                                            count=1,
                                                            freq=MONTHLY,
                                                            byweekday=[SU(-1)]),
                                        time_from = dt.time(23,0),
                                        time_to   = dt.time(23,30),
                                        tz = pytz.timezone("Pacific/Pago_Pago"))
        self.calendar.add_child(instance=self.night)
        self.night.save_revision().publish()

    @timezone.override("Pacific/Kiritimati")
    def testExtremeTZGetAllEventsByDay(self):
        events = getAllEventsByDay(self.request, dt.date(2019,1,1), dt.date(2019,1,31))
        self.assertEqual(len(events), 31)
        evod1 = events[0]
        self.assertEqual(evod1.date, dt.date(2019,1,1))
        self.assertEqual(len(evod1.all_events), 1)
        self.assertEqual(len(evod1.days_events), 1)
        self.assertEqual(evod1.days_events[0].title, "Pacific Night")

# ------------------------------------------------------------------------------
class TestNoCalendar(TestCase):
    def setUp(self):
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('i', 'i@foo.test', 's3cr3t')
        self.request = RequestFactory().get("/test")
        self.request.user = self.user
        self.request.session = {}
        self.request.site = Site.objects.get(is_default_site=True)
        self.group = GroupPage(slug = "initech", title = "Initech Corporation")
        self.home.add_child(instance=self.group)
        self.standup = RecurringEventPage(slug   = "test-meeting",
                                          title  = "Test Meeting",
                                          repeat = Recurrence(dtstart=dt.date(2013,1,1),
                                                              until=dt.date(2013,5,31),
                                                              freq=WEEKLY,
                                                              byweekday=[MO,WE,FR]),
                                          time_from = dt.time(13,30),
                                          time_to   = dt.time(16),
                                          uid       = "initiative+technology")
        self.group.add_child(instance=self.standup)
        self.postponement = PostponementPage(owner = self.user,
                                             slug  = "2013-01-09-postponement",
                                             title = "Postponement for Wednesday 16th of October",
                                             overrides = self.standup,
                                             except_date = dt.date(2013,1,16),
                                             cancellation_title   = "Meeting Postponed",
                                             cancellation_details =
                                                 "The meeting has been postponed until tomorrow",
                                             postponement_title   = "A Meeting",
                                             date      = dt.date(2013,1,17),
                                             time_from = dt.time(13),
                                             time_to   = dt.time(16,30),
                                             details   = "Yes a test meeting on a Thursday")
        self.standup.add_child(instance=self.postponement)

    def testGetAllEventsByDay(self):
        events = getAllEventsByDay(self.request, dt.date(2013,1,1), dt.date(2013,1,31))
        self.assertEqual(len(events), 31)
        evod2 = events[1]
        self.assertEqual(evod2.date, dt.date(2013,1,2))
        self.assertEqual(len(evod2.all_events), 1)
        self.assertEqual(len(evod2.continuing_events), 0)
        self.assertEqual(len(evod2.days_events), 1)

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
